#!/opt/internet/bin/perl
#   #!/usr/local/bin/perl
# ---------------------------------------------------------------------------
$Version = 'iisstat.pl v1.05a';
# $PrintInvalids = 1;	# Print invalid lines from log to STDERR
# $debug = 1;		# debugging on ?
# $PrintWorking = 50;	# Dot every x # of lines
#============================================================================
#  NOTE: This currently only works fully on Perl 4.0
#        5.0 does not execute the Report subroutine correctly
#	 on 5.001e it will run the report sometime and core dump at other times
#============================================================================
#
# Copyright (c) 1995 Steven Nemetz, All Rights Reserved.
#
# This program is provided "AS IS" with absolutely no express or implied
# warranties of any type.
#
# Information on the latest version is at:
# URL: ftp://ftp.ccmail.com/pub/utils/InternetServices/iisstat/iisstat.html
#
# The latest version of iisstat can be found at:
# URL: ftp://ftp.ccmail.com/pub/utils/InternetServices/iisstat/iisstat.tar.gz
#
# To produce graphs with this program, you must have:
#   a) xmgr v3.01 (ftp://ftp.teleport.com/pub/users/pturner/acegr/xmgr-3.01.tar.gz)
#   b) ImageMagick (ftp://ftp.x.org/contrib/applications/ImageMagick)
#   c) ghostscript (ftp://prep.ai.mit.edu/pub/gnu/ghostscript-2.6.2.tar.gz)
#
sub usage {
   chop($ProgName = `basename $0`);
   die <<"EndUsage";
program: Internet Information Services Statistics
usage: $ProgName <options>
$Version

Reports Available:
     -d              include report on domain traffic
     -F              include report by file
     -h              include report on hourly traffic
     -H              include report by hosts
     -m              include report by day of month
     -M              include report by month of year
     -t              include report on total traffic by section
     -T              include report by file types / extensions
     -w              include report by day of week
     -y              include report by date
     -Y              include report by file w/ HTML links & descriptions
     -z	             include report by Organization/domain
                       NOTE: this can be slow ( done by whois lookups )
Input log formats:
     -B              TBBS multiline BBS format log file
     -G	             Gopher format log file
     -N              HTTPd Common Log format log file
     -W              WU-FTPd xfer format log file
     -X              EMWAC HTTP format log file
Output formats:
     -P              output printed text report format (default)
     -C              output Graphs/Charts
     -Z              output HTML report format
Reporting Options:
     -a              include anonymous users ( FTP only )
     -A <reg exp>    report on hosts/addresses matching given Perl reg exp
     -b              block command line options from the config file
     -c <filename>   Configuration file
     -D <domain>     report only on traffic from <domain>
     -E <end date>   specify end date for the report (MMDDYYYY)
                       default - newest date in file
     -f <filename>   Use <filename> for the log file
     -g              include guest users ( FTP only )
     -i              group information files in Section Report
     -I              report Inbound traffic
     -l <depth>      Depth of path detail for sections
     -L	<number>     Top <number> lines of reports
     -n              No report summary
     -o              Report HTML Title override
     -O              report Outbound traffic
     -r              include real users ( FTP only )
     -R              do reverse DNS lookups on numeric IP addresses
                       NOTE: this can be very slow
     -s <reg exp>    report on paths/filenames matching given Perl reg exp 
     -S <start date> specify start date for the report (MMDDYYYY)
                       default - earliest date in file

     Defaults: can be defined in configuration file
             If neither a or r is specified, then a
             If neither O or I is specified, then O
             If neither P or Z is specified, then P
     Default configuration file is: $ConfigFile
EndUsage
}

# Options to add:
#    Reports
#    Output formats:
#       -V              output comma delimited format
#    Input log formats:
# ---------------------------------------------------------------------------

#-------------------------------------------------------------------------
# configuration
# edit the following lines to fit your site
#	some of these can be overwriten by the config file or command line
#-------------------------------------------------------------------------

# Name of the configuration file ONLY
# Will look for this file in same directory as this program
$ConfigFile = "stats.cfg";

#-------------------------------------------------------------------------
#	The following variables can be set in the configuration file
#-------------------------------------------------------------------------
# edit the next two lines to customize for your domain.
# This will allow your domain to be seperated in the domain listing.

$mydom1 = "ccmail";
$mydom2 = "com";

# edit the next few lines to customize for your default log & config files
%UsageFiles = (
        'opt_W', '/var/log/ftp.xferlog',                # WU-FTPd xfer log
        'opt_N', '/usr/local/etc/httpd/logs/access_log',# HTTPd log
        'opt_B', '/home/bfulk/bbs_logs/0195.LOG',       # TBBS log
        'opt_G', '/var/log/gopher.history.log',         # Gopher log
	'opt_X', '/home/snemetz/tmp2/hs950604.log',	# EMWAC log
);

# Document root directory for servers
#	used on Gopher reports to determine bytes transfered &
#	for option -Y
%DataRoots = (
	'opt_W', '/internet/data/ftp',		# FTP's root directory
	'opt_N', '/internet/data/ftp',		# HTTPd's root directory
	'opt_G', '/internet/data/ftp',		# Gopher's root directory
	'opt_X', ' ',	# EMWAC's HTTP root directory
);

# where to put the generated graphs
%GraphDests = (
        'opt_W', '/home/stats/ftp',     # Destination of FTP graphs
        'opt_N', '/home/stats/www',     # Destination of WWW graphs
        'opt_B', '/home/stats/bbs',     # Destination of BBS graphs
        'opt_G', '/home/stats/gopher',  # Destination of Gopher graphs
	'opt_X', '/home/stats/emwac',	# Destination of EMWAC HTTP graphs
);

$HostOverride = 'ftp.ccmail.com';
#-------------------------------------------------------------------------
#	The following variables can NOT be set in the configuration file
#-------------------------------------------------------------------------

# Database of company names found for report -z
$CompanyFile = '/var/log/Companies.dat';
# Database of Host names found with -R
$HostNameFile = '/var/log/HostNames.dat';

# File used to find file descriptions for report -Y
$IndexFile = '00_index';

# Make sure the linker can find the Motif and X libraries,
# if this is run from cron ( needed for graphic programs )
$ENV{'LD_LIBRARY_PATH'} = "/usr/openwin/lib:/opt/SUNWmotif/lib";

# where is grbatch (from xmgr package)
$graphprogram = "/usr/local/bin/grbatch";

# where is convert (from ImageMagick, which requires ghostscript)
$graphconvertor = "/usr/local/bin/convert";

$tmp = "/tmp";          # where should tmp files be written ?
$csize    = "100x100";  # size of the chip image
$ccolor   = 10;         # colors for the chip image
$bgc      = 1;          # background color
$outlinec = 5;          # bar outline color
$fillc    = 2;          # bar fill color

#-------------------------------------------------------------------------
# Lines below here are unlikely to need to be changed
#-------------------------------------------------------------------------

# Common Log httpd status result codes.
# Full specifications for these codes are available at
# http://www.w3.org/hypertext/WWW/Protocols/HTTP/HTRESP.html
%HTTPStatus = (
	# 2xx Success
        '200', ' HTTP:Success - OK',
		# Successful request
	'201', ' HTTP:Success - Created',
		# Successful POST
	'202', ' HTTP:Success - Accepted',
		# Accepted for processing, but processing not completed
	'203', ' HTTP:Success - Partial Information',
		# Information returned may not be complete
	'204', ' HTTP:Success - No Response',
		# Don't display anything, stay in the same document
	# 3xx Redirection
	'301', ' HTTP:Redirection - Moved',
        '302', ' HTTP:Redirection - Found',
	'303', ' HTTP:Redirection - Method',
        '304', ' HTTP:Redirection - Not Modified',
	# 4xx, 5xx Errors
        '400', ' HTTP:Error - Bad Requests',
        '401', ' HTTP:Error - Unauthorized',
	'402', ' HTTP:Error - Payment Required',
        '403', ' HTTP:Error - Forbidden',
        '404', ' HTTP:Error - Not Found',
        '500', ' HTTP:Error - Internal Error',
        '501', ' HTTP:Error - Not Implemented',
	'502', ' HTTP:Error - Service Overloaded',
	'503', ' HTTP:Error - Gateway Timeout'
);
# Estimated bytes sizes of response headers
%HTTPStatusBytes = (
        '200', 180,
        '302', 255,
        '304', 98,
        '400', 313,
        '401', 281,
        '403', 268,
        '404', 250,
        '500', 485,
        '501', 482
);

$opt_l = 3;
# $SearchPath = '';
# $SearchAddress = '';

#-------------------------------------------------------------------------
# NO lines below here should need to be changed
#-------------------------------------------------------------------------
require 'timelocal.pl';
require 'getopts.pl';
require 'sys/socket.ph';

$Wstr = 'SunMonTueWedThuFriSat';
$Mstr = 'JanFebMarAprMayJunJulAugSepOctNovDec';
$PID = "$$";
$xferOFiles = 0; $xferIFiles = 0; $xferOBytes = 0; $xferIBytes = 0;
# $AF_INET = 2; $SOCK_STREAM = 1;
$ConfigFile = &GetConfigFilePath;

$validopts = 'aA:bBc:CdD:E:f:FgGhHiIl:L:mMnNo:OPrRs:S:tTwWXyYzZ';
&Getopts($validopts);

$ConfigFile = $opt_c if ($opt_c);
# Read configuration file
$XAdd = 0; $XFile = 0; $IAdd = 0; $IFile = 0;
open(CONFIG, $ConfigFile) || die "Error opening configuration file: $ConfigFile\n";
cline: while(<CONFIG>) {
       @line = split;
       # Get domain names
       # Format in config file:
       #	DOMAIN domain description
       if ($line[0] eq "DOMAIN") {
         $line[1] =~ tr/A-Z/a-z/;
	 $DomainName{$line[1]} = join(" ",@line[2..$#line]); }
       # Get address patterns to be excluded from the report
       elsif ($line[0] eq "EXCLUDEADDRESS") {
         $line[1] =~ s/([#])/\\\1/g;    # escape # characters
         $ExcludeAdd[$XAdd++] = $line[1];
       }
       # Get file/path patterns to be excluded from the report
       elsif ($line[0] eq "EXCLUDEFILE") {
         $line[1] =~ s/([#])/\\\1/g;    # escape # characters
         $ExcludeFiles[$XFile++] = $line[1];
       }
       # Get address patterns to be included in the report
       elsif ($line[0] eq "INCLUDEADDRESS") {
         $line[1] =~ s/([#])/\\\1/g;    # escape # characters
         $IncludeAdd[$IAdd++] = $line[1];
       }
       # Get file/path patterns to be included in the report
       elsif ($line[0] eq "INCLUDEFILE") {
         $line[1] =~ s/([#])/\\\1/g;    # escape # characters
         $IncludeFiles[$IFile++] = $line[1];
       }
       # get whois servers
       # Format in config file:
       #	WHOIS DomainPattern WhoisServer "Query" OrganizationPattern
       elsif ($line[0] eq "WHOIS") {
	 ($dp, $serv, $query, $op) = /^\s*WHOIS\s+(\S+)\s+(\S+)\s+"(.+)"\s+(.*)$/;
	 $Whois{$dp} = join($;,($serv, $query, $op) );
       }
       # Get log locations
       elsif ($line[0] eq "FTPXFERLOG")    { $UsageFiles{'opt_W'} = $line[1]; }
       elsif ($line[0] eq "BBSLOG")        { $UsageFiles{'opt_B'} = $line[1]; }
       elsif ($line[0] eq "HTTPLOG")       { $UsageFiles{'opt_N'} = $line[1]; }
       elsif ($line[0] eq "GOPHERLOG")     { $UsageFiles{'opt_G'} = $line[1]; }
       elsif ($line[0] eq "EMWACHTTPLOG")  { $UsageFiles{'opt_X'} = $line[1]; }
       # Get graph destinations
       elsif ($line[0] eq "FTPGRAPH")      { $GraphDests{'opt_W'} = $line[1]; }
       elsif ($line[0] eq "BBSGRAPH")      { $GraphDests{'opt_B'} = $line[1]; }
       elsif ($line[0] eq "HTTPGRAPH")     { $GraphDests{'opt_N'} = $line[1]; }
       elsif ($line[0] eq "GOPHERGRAPH")   { $GraphDests{'opt_G'} = $line[1]; }
       elsif ($line[0] eq "EMWACHTTPGRAPH"){ $GraphDests{'opt_X'} = $line[1]; }
       # Get root directories of servers
       elsif ($line[0] eq "FTPROOT")       { $DataRoots{'opt_W'} = $line[1]; }
       elsif ($line[0] eq "HTTPROOT")      { $DataRoots{'opt_N'} = $line[1]; }
       elsif ($line[0] eq "GOPHERROOT")    { $DataRoots{'opt_G'} = $line[1]; }
       elsif ($line[0] eq "EMWACHTTPROOT") { $DataRoots{'opt_X'} = $line[1]; }
       # Get local domain name
       elsif ($line[0] eq "MYDOMAIN") {
         ($mydom1, $mydom2) = split(/\./,$line[1]); }
       elsif ($line[0] eq "HOSTOVERRIDE") { $HostOverride = $line[1]; }
       elsif (!$opt_b && $line[0] eq "IISTATOPTS") {
         local(@CLineOpts) = @ARGV;
         @ARGV = @line[1..$#line];
         &Getopts($validopts);
         @ARGV = @CLineOpts;
         &Getopts($validopts);
       }
}
close CONFIG;

# display usage if no log format
&usage if (!$opt_W && !$opt_N && !$opt_G && !$opt_X && !$opt_B);
if ($opt_B && ($opt_d || $opt_H || $opt_t || $opt_Y || $opt_z || $opt_R)) {
        print "ERROR: -d -H -R -t -Y and -z cannot be used with -B\n";
        &usage; }
if ($opt_Y && ! $opt_Z) {
	print "ERROR: -Y can only be used with -Z\n";
	&usage; }
if ($opt_Y && $opt_X ) {
	print "ERROR: -Y is not supported with -X\n";
	$usage; }
# Get the log file name
type: foreach $logtype (keys %UsageFiles) {
   if (eval "defined(\$$logtype);") {
      if ($found) {
         print STDERR "More than one log format type selected\n";
         print STDERR "Only one log is currently supported at a time\n";
         exit;
      }
      $usage_file = $UsageFiles{$logtype};
      $found = 1;
      next type;
   }
}
undef $found;
# Get the graph destination path
dest: foreach $graphpath (keys %GraphDests) {
   if (eval "defined(\$$graphpath);") {
      if ($found) {
         print STDERR "More than one graph path selected\n";
         print STDERR "Only one log is currently supported at a time\n";
         exit;
      }
      $destination = $GraphDests{$graphpath};
      $found = 1;
      next dest;
   }
}
undef $found;
 
$anon  = 1 if ($opt_a);
$guest = 1 if ($opt_g);
$real  = 1 if ($opt_r);
$anon  = 1 if (!$opt_a && !$opt_r && !$opt_g);
$usage_file = $opt_f if ($opt_f);
$opt_t = 1 if ($opt_i);
$opt_O = 1 if (!$opt_O && !$opt_I);
$opt_P = 1 if (!$opt_P && !$opt_Z);
$SDate = &OptToDate($opt_S) if ($opt_S);
$EDate = &OptToDate($opt_E) if ($opt_E);
$title = $opt_o if $opt_o;
if ($opt_A) {
        $SearchAddress = $opt_A;
        $SearchAddress =~ s/([#.])/\\\1/g;      # Escape # and . characters
}
if ($opt_s) {
        $SearchPath = $opt_s;
        $SearchPath =~ s/([#.])/\\\1/g;         # Escape # and . characters
}

if ($opt_z && -e $CompanyFile) {	# read in the Companies file
   open(DATAFILE, "$CompanyFile" ) || die "Error: $!";
   while (<DATAFILE>) {
      @line = split;
      if ($Companies{$line[0]}) {	# Duplicate found in file
	 $Dup++;
      }
      else { $Companies{$line[0]} = join(' ',@line[1..$#line]); }
   }
   close(DATAFILE);
   if ($Dup) {	# Write file back out to clean up duplicates
      open(DATAFILE, ">$CompanyFile") || die "Error: $!";
      foreach $item ( sort keys(%Companies)) {
	 print DATAFILE $item, ' ' x (20 - length($item)), " ",
			$Companies{$item}, "\n";
      }
      close(DATAFILE);
   }
   undef( $Dup );
}
if ($opt_R && -e $HostNameFile) {	# read in the Hosts file
   open(DATAFILE, "$HostNameFile" ) || die "Error: $!";
   while (<DATAFILE>) {
      @line = split;
      if ($HostNames{$line[0]}) {	# Duplicate found in file
	 $Dup++;
      }
      else { $HostNames{$line[0]} = join(' ',@line[1..$#line]); }
   }
   close(DATAFILE);
   if ($Dup) {	# Write file back out to clean up duplicates
      # should sort be number
      open(DATAFILE, ">$HostNameFile") || die "Error: $!";
      foreach $item ( sort keys(%HostNames)) {
	 print DATAFILE $item, ' ' x (17 - length($item)),
			$HostNames{$item}, "\n";
      }
      close(DATAFILE);
   }
   undef( $Dup );
}

($userID) = getpwuid($<);	# Remove for Windows NT
if ($HostOverride) { $hostname = $HostOverride; }
else { $hostname = `hostname`; chop($hostname); }
$subtitle = "$hostname";

# Sufixes of statistic variables
%VarCodes = ( 'd', 'Domain',
              'F', 'File',
              'h', 'Hour',
              'H', 'Host',
              'm', 'MDay',
              'M', 'Month',
              't', 'Section',
              'T', 'Type',
              'w', 'Weekday',
              'y', 'Date',
	      'Y', 'File',
	      'z', 'Company'
);
# Names of comparision routines for sorting
%SortCodes = ( 'd', 'CompareFileDomain',# sort by # files, domain
               'F', 'CompareFileAlpha', # sort by # files, file
               'h', 'CompareNumber',    # sort by hour
               'H', 'CompareFileHost',  # sort by # files, host w/ IP#s last
               'm', 'CompareNumber',    # sort by day number
               'M', 'CompareMonth',     # sort to Jan - Dec
               't', 'CompareFileAlpha', # sort by # files, section
               'T', 'CompareFileAlpha', # sort by # files, type of file
               'w', 'CompareWeekday',   # sort to Sun - Sat
               'y', 'CompareRDate',     # sort by date, newest first
	       'Y', 'CompareFileAlpha',	# same as 'F'
	       'z', 'CompareFileAlpha'	# sort by # files, Company name
);

# TO DO: CREATE main loop to handle multiple files

open (LOG,$usage_file) || die "Error opening usage log file: $usage_file\n";


# Process the log file
line: while (<LOG>) {

   # Split data from different log types to fields
   #    and skip any line that is not being reported on

   if ($opt_W) {        # process WU-FTPd line
      @line = split;
      # skip if WU-FTPd record is one we are not interested in
      if ($#line != 16) {                        # skip if not 17 fields
         print STDERR "Invalid line: $_" if ($PrintInvalids);
         next line;
      }
      next line if (!$anon && $line[12] eq "a"); # skip if not reporting anonymous users
      next line if (!$real && $line[12] eq "r"); # skip if not reporting real users
      next line if (!$guest && $line[12] eq "g"); # skip if not reporting guests 
      # split WU-FTPd log record
      $wday     = $line[0];                # set week day
      $date = join(' ', @line[1,2,4]);     # set the date ( Month Day Year )
      $hour     = substr($_,11,2);         # set the hour
      $xfersec  = $line[5];                # set number of secs the transfer took
      $addr     = $line[6];                # set host
      $bytes    = $line[7];                # set size of file in bytes
      $filename = $line[8];                # set file name
      $direct   = $line[11];               # inbound/outbound transfer
   }
   elsif ($opt_N) {     # process HTTPd Common Log Format line ( NCSA, CERN )
      # parse line
      ($addr, $rfc931, $authuser, $timestamp, $request, $status, $bytes) =
         /^(\S+) (\S+) (\S+) \[(.+)\] \"(.+)\" (\S+) (\S+)\s/;
      # verify that it is a valid line
      if (!($addr && $rfc931 && $authuser && $timestamp && $request && $status)) {
         print STDERR "Invalid line: $_" if ($PrintInvalids);
         next line;
      }
      $bytes = 0 if ($bytes eq '-');
      $bytes += $HTTPStatusBytes{$status};
      if ($status != 200) {      # an error
#        print STDERR "Error line: $_" if ($debug);
        $FilesType{$HTTPStatus{$status},'o'}++;
        $BytesType{$HTTPStatus{$status},'o'} += $bytes;
        next line;
      }
      $xfersec = 'N/A';
      # Process date and time
      @timestamp = split(/[\s\/:]/, $timestamp);
      $date = join(' ',@timestamp[1,0,2]);
      $hour = $timestamp[3];
      $timestamp[2] = substr($timestamp[2],2,2);
      $wday = &WeekDay(@timestamp[1,0,2,3,4,5,6]);
      # Process request
      @request = split(' ',$request);
      if ($#request >= 1 ) {            # has method and filename
         $filename = $request[1];       # Valid methods: GET, HEAD, POST
         $method = $request[0];
      }
      else {
         $filename = $request[0];
         $method = 'GET';
      }
      $direct = 'o';
#      if ($method eq 'GET') { $direct = 'o'; }
#      else { $direct = 'i'; }
   }
   elsif ($opt_X) {	# process EMWAC HTTP line
      @line = split;
      if ($#line < 8 ) {
	 print STDERR "Invalid line: $_" if ($PrintInvalids);
	 next line;
      }
      $wday = $line[0];                # set week day
      $date = join(' ', @line[1,2,4]); # set the date ( Month Day Year )
      $hour = substr($_,11,2);         # set the hour
      $addr = $line[6];
      $filename = $line[8];
      $bytes = 0;
      $xfersec = 'N/A';
      $direct = 'o';
   }
   elsif ($opt_G) {	# process gopher line
      ($wday, $mon, $day, $hour, $year, $addr, $message) =
      /^(\S+)\s+(\S+)\s+(\S+)\s+([0-9]+):\S+\s+([0-9]+)\s+\S+\s+(\S+) : (.*)/;
      @message = split(' ', $message);
      # verify that it is a valid line
      if ( !($wday && $mon && $day && $hour && $year && $addr && $message)
	   || $message[1] eq "directory" 	# Skip directories
           || $message[1] eq "Connection"	# Skip directories
	   || $message[1] eq "Timed"		# Skip time outs
	   || $message[0] eq "Denied"		# Skip denied accesses
	   || $message[0] eq "Client"		# Skip "Client went away!"
	 ) {
	 print STDERR "Invalid line: $_" if ($PrintInvalids);
	 next line;
      }
      $date = join(' ',($mon, $day, $year));
      $filename = $message[$#message];
      if ($debug) {
         print STDERR "Message: $message\n";
	 print STDERR "  Filename: $filename\n";
      }
#      $filename =~ s#^\.\.##;
      $bytes = (-s "$DataRoots{'opt_G'}/$filename");
#      $filename =~ s#^[/\.]+#/#;
#      $filename =~ s#//#/#;
      $direct = 'o';
      $xfersec = 'N/A';
   }
   elsif ($opt_B) {     # process TBBS line
      undef ($line, $date, $time, $protocol, $direction, $filename, $other, $other2);
      # parse line
      ($line, $date, $time, $protocol, $direction, $filename, $other, $other2) = split;
      if ($line !~ /^\([0-9]{1,2}\)$/ ||        # Skip if no line number
          $other2 ||                            # Skip if 8 or more fields
          $protocol eq 'Message' ||
          $protocol eq '***' ||
          $protocol eq 'Logged' ||
          $protocol eq 'Loaded' ||
          !($line && $date && $time && $protocol && $direction) ) { # Skip if under 5 fields
                print STDERR "Invalid line: $_" if ($PrintInvalids);
                next line;
      }
#      print STDERR "Good line: $_" if ($debug);
      # Line could have 5-7 fields
      if ($other && $filename eq '(Aborted)') {         # ( 7 fields )
        # aborted downloads ( 7 fields )
        $FilesType{' BBS:User Aborted Transfers',$direction eq 'Download'?'o':'i'}++;
        $BytesType{' BBS:User Aborted Transfers',$direction eq 'Download'?'o':'i'} = 0;
        next line;      # $filename = $other;
      }
      elsif ($protocol eq 'Read') {                     # ( 5 fields )
        $filename = $direction;
        $direction = 'Download'; }
      elsif ($protocol eq 'Uploaded') {                 # ( 5 fields )
        $filename = $direction;
        $direction = 'Upload'; }
      # else has 6 fields as named
      @time = split(/:/, $time);
      $hour = $time[0];
      @date = split(/-/,$date);
      $date[0] = substr($Mstr, ($date[0]-1)*3, 3);
      next line if ($#date != 2);
      $wday = &WeekDay(@date, @time, 0);
#      $date[2] += 1900;
      $date = join(' ',@date);
      # strip path, only the filename is desired
      @path = split(/\\/, $filename);
      $filename = $path[$#path];
#      print STDERR "ERROR Line: $_" if ($debug && $filename !~ /.+\..{0,3}$/);
      if ($direction eq 'Download') { $direct = 'o'; }
      else { $direct = 'i'; }
      $xfersec = 'N/A';
      $addr = 'N/A';
      $bytes = 0;
   }
   else { die "No valid log type found\n"; }

   # skip if date is not in the reporting range
   next line if ($opt_S && &CompareDate($SDate, $date) > 0);
   next line if ($opt_E && &CompareDate($date, $EDate) > 0);

   # Handle file/path inclusions - skip if path does not match
   if ($SearchPath) {next line unless ($filename =~ m#$SearchPath#);}
   # Handle file/path inclusion lists  - skip if path does not match
   if ($#IncludeFiles >= 0) {
      $found = 0;
      foreach $index (@IncludeFiles) {
         $found = 1 if ($filename =~ m#$index#);
      }
      next line if ($found == 0);
   }
   # Handle file/path exclusions lists - skip is file is being excluded
   foreach $index (@ExcludeFiles) { next line if ($filename =~ m#$index#); }

#   print STDERR "Filename: $filename\n" if ($debug);

   # set the archive area
   if ($filename eq "\.") { $filename = "/unreadable/filename";}
   @path = split(/\//, $filename);
   # Things in the top-level directory and all . files
   #   are assumed to be informational files
   if ($#path <= 1 || ( $opt_i && substr($path[$#path],0,1) eq "."))
      { $pathkey = "Index/Informational Files"; }
   elsif ( substr($filename, -1) eq '/' ) {
      # Handle case when the request was the default document
      # Request ended with a /
      $pathkey=join("/",@path[0..($#path < $opt_l ? $#path : $opt_l)]);
   }
   else {
      $pathkey=join("/",@path[0..(($#path - 1) < $opt_l ? $#path -1 : $opt_l)]);
   }
   undef @path;

   # set the domain & host addresses
   if ($opt_R && $addr =~ /^[0-9\.]+$/) {
      if ( $HostNames{$addr} ) { $addr = $HostNames{$addr}; }
      else {
	 $HostNames{$addr} = &hostbyaddr($addr);
	 #  save $HostNames table to disk to speed up future reports
	 open(DATAFILE, ">>$HostNameFile") || die "Error: $!";
	 print DATAFILE $addr, ' ' x (17 - length($addr)),
			$HostNames{$addr}, "\n";
	 close(DATAFILE);
	 $addr = $HostNames{$addr};
      }
   }
   $addr =~ tr/A-Z/a-z/;
   @address = split(/\./, $addr);
   $domain = $address[$#address];       # assign top domain
   if (($domain eq "$mydom2" && $address[$#address-1] eq "$mydom1")
        || $#address == 0) {
      # Address in local domain
      $domain = "$mydom1\.$mydom2";
      $addr .= "\.$mydom1\.$mydom2" if ($#address == 0);
   }
   # TO DO: add code to check address as well as host name
   # Handle address/host inclusions - skip if address does not match
   if ($SearchAddress) {next line unless ($addr =~ m#$SearchAddress#);}
   # Handle address/host inclusion lists  - skip if address does not match
   if ($#IncludeAdd >= 0) {
      $found = 0;
      foreach $index (@IncludeAdd) {
         $found = 1 if ($addr =~ m#$index#);
      }
      next line if ($found == 0);
   }
   # Handle address/host exclusion lists - skip if address is being excluded
   foreach $index (@ExcludeAdd) { next line if ($addr =~ m#$index#); }

   if ( $#address == 3 && $address[3] =~ /^[0-9]+$/ ) {
      # Numeric IP address
      # Do not reverse IP addresses
      $domain = "unresolved";
      $Host = $addr; }
   else {
      # Host name
      # Reverse the address, so we sort by reverse subdomain on hosts
      $Host = join('.', reverse(split(/\./, $addr)));
   }
   undef @address;

   # Skip if reporting on a single domain and the address is not in it
   next line if ($opt_D && substr($addr,length($addr) - length($opt_D),
                           length("$opt_D")) ne "$opt_D" );

   # look up the company
   if ( $opt_z ) {
      if ( $domain eq "unresolved" ) { $Company = "unresolved"; }
      else {
	 @Dom = split(/\./, $Host);
	 $#Dom = 2 if $#Dom > 2;	# reduce to last 3 parts of the name
         if ($Dom[0] =~ /^(arpa|com|edu|gov|net|org)$/ && $#Dom == 2 ) { pop(@Dom); }
         $DomAddr = join('.', @Dom);
         if ( $Companies{$DomAddr} ) { $Company = $Companies{$DomAddr}; }
	 elsif ( $#Dom == 2 && $Companies{join('.',@Dom[0..1])} ) {
	    $Company = $Companies{join('.',@Dom[0..1])}; }
         else {		# Try to get new information
	    $Company = $Companies{$DomAddr} = &GetCompany($DomAddr);
	    if ( $#Dom == 2 && $Company eq $DomAddr ) {	# Try again w/ 2 part
	      pop(@Dom); delete( $Companies{$DomAddr} ); # or undef
              $DomAddr = join('.', @Dom);
              if ( $Companies{$DomAddr} ) { $Company = $Companies{$DomAddr}; }
	      else {
		 $Company = $Companies{$DomAddr} = &GetCompany($DomAddr);
		 $NewComp++;
	       }
	    }
	    else { $NewComp++; }
	    if ( $NewComp ) {
	       # save new entry to disk to speed up future reports
	       open(DATAFILE, ">>$CompanyFile") || die "Error: $!";
	       print DATAFILE $DomAddr, ' ' x (20 - length($DomAddr)), " ",
			      $Companies{$DomAddr}, "\n";
	       close(DATAFILE);
	    }
	 }
      }
      undef( $NewComp, @Dom, $DomAddr );
   }
   
   # Let user know we are still working
   if ( $PrintWorking ) {
      if ( $WorkingCount == $PrintWorking ) {
	 print STDERR ".";
	 $PrintedDots++;
	 $WorkingCount = 0;
	 if ( $PrintedDots == 70 ) {
	    print STDERR "\n";
	    $PrintedDots = 0;
	 }
      }
      $WorkingCount++;
   }
   # add to the statistics
   if ($direct eq 'o') {
     # Totals Statistics of files sent
     $xferOFiles++;                         # total files sent
     $xferOBytes          += $bytes;        # total bytes sent
   }
   else {
     # Totals Statistics of files received
     $xferIFiles++;                         # total files received
     $xferIBytes          += $bytes;        # total bytes received
   }

   # Only create data for needed reports
#   if ($direct eq 'o') {
#      $FilesDateo{$date}++
#      ...
#   }
   $FilesDate{$date,$direct}++;               # files per date
   $FilesHost{$Host,$direct}++;               # files per host address
   if ($opt_y) {
      $BytesDate{$date,$direct}     += $bytes;   # bytes per date
      $SecsDate{$date,$direct}      += $xfersec; # per date
   }
   if ($opt_d) {
      $FilesDomain{$domain,$direct}++;             # files per domain
      $BytesDomain{$domain,$direct}   += $bytes;   # bytes per domain
      $SecsDomain{$domain,$direct}    += $xfersec; # per domain
   }
   if ($opt_F || $opt_T || $opt_Y) {
      $FilesFile{$filename,$direct}++;           # files per file
      $BytesFile{$filename,$direct} += $bytes;   # bytes per file
   }
   if ($opt_t) {
      $FilesSection{$pathkey,$direct}++;            # files per archive area
      $BytesSection{$pathkey,$direct}  += $bytes;   # bytes per archive area
#      $SecsSection{$pathkey,$direct}   += $xfersec; # per archive area
   }
   if ($opt_H) {
      $BytesHost{$Host,$direct}     += $bytes;   # bytes per host address
      $SecsHost{$Host,$direct}      += $xfersec; # per host address
   }
   if ($opt_h) {
      $FilesHour{$hour,$direct}++;               # files per hour of day
      $BytesHour{$hour,$direct}     += $bytes;   # bytes per hour of day
      $SecsHour{$hour,$direct}      += $xfersec; # per hour of day
   }
   if ($opt_m) {
      $FilesMDay{(split(' ',$date))[1],$direct}++; # files per day of month
      $BytesMDay{(split(' ',$date))[1],$direct} += $bytes; # bytes per day of month
      $SecsMDay{(split(' ',$date))[1],$direct} += $xfersec; # per day of month
   }
   if ($opt_M) {
      $FilesMonth{(split(' ',$date))[0],$direct}++;    # files per month
      $BytesMonth{(split(' ',$date))[0],$direct} += $bytes;  # bytes per month
      $SecsMonth{(split(' ',$date))[0],$direct} += $xfersec; # per month
   }
   if ($opt_w) {
      $FilesWeekday{$wday,$direct}++;               # files per day of week
      $BytesWeekday{$wday,$direct}     += $bytes;   # bytes per day of week
      $SecsWeekday{$wday,$direct}      += $xfersec; # per day of week
   }
   if ($opt_z) {
      $FilesCompany{$Company,$direct}++;               # files per Company
      $BytesCompany{$Company,$direct}     += $bytes;   # bytes per Company
      $SecsCompany{$Company,$direct}      += $xfersec; # per Company
   }
}	# end of while (<LOG>)
close LOG;

if ($opt_O && $xferOFiles == 0)	{
   die "There was no outbound traffic data to process.\n";}
if ($opt_I && $xferIFiles == 0) {
   die "There was no inbound traffic data to process.\n";}

# calculate data for rest of reports
&CreateTypeData if ($opt_T);

# Create Reports
$= = 59;
if ($opt_Z) {   # HTML report header
   # TO DO: add code to put what type of server in title
   $title = "$hostname Server Statistics" if ! $title;
   print "<HTML><HEAD>\n";
   print "<TITLE>$title</TITLE>\n";
   print "</HEAD><BODY>\n";
   print "<H1>$title</H1>\n";
}

if ($opt_D) {print "Transfer Totals include the '$opt_D' domain only.\n";
	     print "All other domains are filtered out for this report.\n\n";}
if ($opt_s) {print "Transfer Totals include paths matching '$SearchPath' only.\n";
	     print "All other files are filtered out for this report.\n\n";}
if ($opt_A) {print "Transfer Totals include hosts matching '$SearchAddress' only.\n";
	     print "All other hosts are filtered out for this report.\n\n";}

# print message to let user know only certain files are included in this report
if ($#IncludeFiles >= 0) {
   print "Transfer Totals include the following file path patterns only.\n";
   $linelength = 0;
   for ($i = 0; $i <= $#IncludeFiles; $i++) {
      $linelength =+ length($IncludeFiles[$i]) + 2;
      if ( $linelength > 80 ) {
         print "\n";
	 $linelength = length($IncludeFiles[$i]) + 2;
      };
      print $IncludeFiles[$i];
      print ", " if ( $i != $#IncludeFiles );
   }
   print "\nAll other files are filtered out for this report.\n\n";
}

# print message to let user know only certain addresses are included
# in this report
if ($#IncludeAdd >= 0) {
   print "Transfer Totals include the following address patterns only.\n";
   $linelength = 0;
   for ($i = 0; $i <= $#IncludeAdd; $i++) {
      $linelength =+ length($IncludeAdd[$i]) + 2;
      if ( $linelength > 80 ) {
         print "\n";
	 $linelength = length($IncludeAdd[$i]) + 2;
      };
      print $IncludeAdd[$i];
      print ", " if ( $i != $#IncludeAdd );
   }
   print "\nAll other addresses are filtered out for this report.\n\n";
}

# add routines to print out what is excluded from the report

&ReportTotals if (!$opt_n);
if ($opt_Z && 		    	# HTML HREFs to sections in report 
    ($opt_d + $opt_F + $opt_h +	# ONLY display refs if more than 1 report
     $opt_H + $opt_m + $opt_M +
     $opt_t + $opt_T + $opt_w +
     $opt_y + $opt_Y + $opt_z) > 1) {
   # HREF Anchors have the same names as the graph files
   # TO DO: Add code to handle both In and Out reports
   #       anchor = $filename$direction
   print "Statistic Reports Available:\n";
   print "<UL>\n";
   print "<LI><A HREF=\"#date\">Daily Transmission Statistics</A>\n" if ($opt_y);
   print "<LI><A HREF=\"#hour\">Hourly Transmission Statistics</A>\n" if ($opt_h);
   print "<LI><A HREF=\"#wday\">Weekday Transmission Statistics</A>\n" if ($opt_w);
   print "<LI><A HREF=\"#mday\">Day of Month Transmission Statistics</A>\n" if ($opt_m);
   print "<LI><A HREF=\"#month\">Monthly Transmission Statistics</A>\n" if ($opt_M);
   print "<LI><A HREF=\"#domain\">Total Transfers by Domain</A>\n" if ($opt_d);
   print "<LI><A HREF=\"#company\">Total Transfers by Organization</A>\n" if ($opt_z);
   print "<LI><A HREF=\"#type\">Total Transfers by File Type</A>\n" if ($opt_T);
   print "<LI><A HREF=\"#archive\">Total Transfers from each Archive Section</A>\n" if ($opt_t);
   print "<LI><A HREF=\"#TopFiles\">Top $opt_L requested files</A>\n" if ($opt_Y);
   print "<LI><A HREF=\"#file\">Total Transfers of each file</A>\n" if ($opt_F);
   print "<LI><A HREF=\"#host\">Client Transmission Statistics</A>\n" if ($opt_H);
   print "</UL>\n";
# put in ref for graphs somewhere
}
$xferOBytes = 1 if ($xferOBytes < 1);
$xferIBytes = 1 if ($xferIBytes < 1);
&CreateReports('o') if ($opt_O);	# Create outbound traffic reports
&CreateReports('i') if ($opt_I);	# Create inbound traffic reports
if ($opt_Z) {   # HTML footer
   print "<P><HR><ADDRESS>Report produced by ";
   print "<A HREF=\"ftp://ftp.support.lotus.com/";
   print "pub/utils/InternetServices/iisstat/iisstat.html\">";
   print "$Version</A></ADDRESS>\n";
   print "</BODY></HTML>\n";
}
exit(0);

#-------------------------------------------------------------------------
# All subroutines are below this point
#-------------------------------------------------------------------------

#-------------------------------------------------------------------------
#	Report subroutines
#-------------------------------------------------------------------------
sub CreateReports {
  local($direction) = $_[0];
  &Report('Daily Transmission Statistics',
          ' ', 'Date', 'y', $direction,
          'y', 'date', 'Last 7 Days Traffic Summary', 'bar',
          'Day', 'Requests', -1, 7, 0, 10, 'y', 7)              if ($opt_y);
  &Report('Hourly Transmission Statistics',
          ' ', 'Time', 'h', $direction,
          'y', 'hour', 'Hourly Traffic Summary', 'bar',
          'Time', 'Requests', -1, 24, 0, 10, 'y', 24)           if ($opt_h);
  &Report('Weekday Transmission Statistics',
          ' ', 'Weekday', 'w', $direction,
          'y', 'wday', 'Weekday Traffic Summary', 'bar',
          'Day', 'Requests', -1, 7, 0, 10, 'y', 8)              if ($opt_w);
  &Report('Day of Month Transmission Statistics',
          ' ', 'Day', 'm', $direction,
          'y', 'mday', 'Day of Month Traffic Summary', 'bar',
          'Day', 'Requests', -1, '$count', 0, 10, 'y', 32)      if ($opt_m);
  &Report('Monthly Transmission Statistics',
          ' ', 'Month', 'M', $direction,
          'y', 'month', 'Monthly Traffic Summary', 'bar',
          'Month', 'Requests', -1, 12, 0, 10, 'y', 12)          if ($opt_M);
  &Report('Total Transfers by Domain',
          ' ', 'Domain Name', 'd', $direction,
          'y', 'domain', 'Domain Traffic Summary', 'hbar',
          'Requests', 'Domain', 0, 1000, -1, '$count', 'x', 40) if ($opt_d);
  &Report('Total Transfers by Organization',
          ' ', 'Organization Name', 'z', $direction,
          'y', 'company', 'Organization Traffic Summary', 'hbar',
          'Requests', 'Organization', 0, 1000, -1, '$count', 'x', 40) if ($opt_z);
  &Report('Total Transfers by File Type',
          ' ', 'File Type', 'T', $direction,
          'n', 'type', 'File Type Traffic Summary', 'hbar',
          'Requests', 'File Type', 0, 1000, -1, '$count', 'x', 40) if ($opt_T);
  &Report('Total Transfers from each Archive Section',
          ' ', 'Archive Section', 't', $direction,
          'n', 'archive', 'Archive Section Traffic Summary', 'hbar',
          'Request', 'Archive', 0, 1000, -1, '$count', 'x', 40) if ($opt_t);
  &Report("Top $opt_L requested files",
          ' ', ' ', 'Y', $direction,
          ' ', 'TopFiles', 'File Traffic Summary', 'hbar',
          'Requests', 'File', 0, 1000, -1, '$count', 'x', 40)   if ($opt_Y);
  &Report('Total Transfers of each file',
          ' ', 'File', 'F', $direction,
          'n', 'file', 'File Traffic Summary', 'hbar',
          'Requests', 'File', 0, 1000, -1, '$count', 'x', 40)   if ($opt_F);
  &Report('Client Transmission Statistics',
          'Hosts by', 'Reversed Subdomain', 'H', $direction,
          'y', 'host', 'Client Traffic Summary', 'hbar',
          'Requests', 'Host', 0, 1000, -1, '$count', 'x', 40)   if ($opt_H);
}

sub ReportTotals {
  # Report Totals
  local (%systems, %dateonly, @syslist, @dates);
  # need to calc number of days in range instead of only ones that have data for
  foreach $date (keys(%FilesDate)) {
        $dateonly{(split($;, $date))[0]}++;
#        @date = split($;, $date);
#        $dateonly{$date[0]}++;
  }
  @dates = sort ByDate keys(%dateonly);
  $subtitle = "$subtitle - $dates[$#dates] to $dates[0]";
  foreach $host (keys(%FilesHost)) {
        $systems{(split($;, $host))[0]}++;
#        @host = split($;, $host);
#        $systems{$host[0]}++;
  }
  @syslist = keys(systems);
  if ($opt_Z) { # HTML
     print "<PRE>\n";
     print "<B>";
  }
  print "TOTALS FOR SUMMARY PERIOD ", $dates[$#dates], " TO ", $dates[0];
  print $opt_Z ? "</B>" : "", "\n\n";
  printf("Files Transmitted During Summary Period       %12.0f\n", $xferOFiles);
  printf("Files Received During Summary Period          %12.0f\n", $xferIFiles);
  printf("Total Files Transferred During Summary Period %12.0f\n",
	$xferIFiles + $xferOFiles);
  printf("Bytes Transmitted During Summary Period       %12.0f\n", $xferOBytes);
  printf("Bytes Received During Summary Period          %12.0f\n", $xferIBytes);
  printf("Total Bytes Transferred During Summary Period %12.0f\n",
	$xferIBytes + $xferOBytes);
  printf("Systems Using Archives                        %12.0f\n\n",
      $#syslist+1);

  printf("Average Files Transmitted Daily               %12.0f\n",
      $xferOFiles / ($#dates + 1));
  printf("Average Files Received Daily                  %12.0f\n",
	$xferIFiles / ($#dates + 1));
  printf("Average Bytes Transmitted Daily               %12.0f\n",
      $xferOBytes / ($#dates + 1));
  printf("Average Bytes Received Daily                  %12.0f\n",
	$xferIBytes / ($#dates + 1));
  if ($opt_Z) {
     print "</PRE>\n";
  }
}

sub Report {
   # Generic Reporting routine
   # Parameters: Report Title, Report Subject Line 1, Line 2, Subject Code,
   #             xfer Direction, Display Xmit,
   #             File Name, Graph Title, Graph Type, X Label, Y Label,
   #             Min X, Max X, Min Y, Max Y, Autoscale, Max Data
   local ($ReportTitle, $subject1, $subject2, $code, $direction, $xmit,
          $filename, $GraphTitle, $GraphType, $XLabel, $YLabel,
          $MinX, $MaxX, $MinY, $MaxY, $autoscale, $MaxData ) = @_;
   local ($date, $name, $bytes, $dir);
format ReportTopXmit =

@||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$ReportTitle

@||||||||||||||||||||| ------Number  of------  Average   ----Percent  of----
$subject1
@||||||||||||||||||||| Files @>>> Bytes  @>>> Xmit  Rate File @>>> Byte @>>>
$subject2,                   $dir,       $dir,               $dir,      $dir
---------------------- ---------- ----------- ---------- --------- ---------
.

format ReportLineXmit =
@<<<<<<<<<<<<<<<<<<<<< @>>>>>>>>> @>>>>>>>>>> @>>>>>>>>> @>>>>>>>  @>>>>>>>  
$name,                 $files,    $bytes,     $avgrate,  $pctfiles, $pctbytes
.

format ReportTop =

@||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$ReportTitle

@||||||||||||||||||||||||||||||| ------Number  of------ -----Percent  of-----
$subject1
@||||||||||||||||||||||||||||||| Files @>>> Bytes  @>>> Files @>>> Bytes @>>>
$subject2,                             $dir,       $dir,      $dir,      $dir
-------------------------------- ---------- ----------- ---------- ----------
.

format ReportLine =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @>>>>>>>>> @>>>>>>>>>> @>>>>>>>   @>>>>>>>
$name,                           $files,    $bytes,     $pctfiles, $pctbytes
.

   $xmit = ($xmit eq 'y') && $opt_W;
   $| = 1;      # do not buffer output
   $- = 0;
   $^ = $xmit ? ReportTopXmit : ReportTop;
   $~ = $xmit ? ReportLineXmit : ReportLine;
   $dir = $direction eq "o" ? 'Sent' : 'Recv';
   if ($opt_Z) {        # HTML section header
     print "<HR>\n";
     print "<A NAME=\"$filename\"><H2>$ReportTitle</H2></A>\n";
     print "<PRE>\n";
     if ($code eq 'Y') {
       print " Percent     Filename", ' ' x 18, "Description\n";
       print '-' x 8, ' ', '-' x 16, ' ', '-' x 40, "\n";
     }
     else {
       if ($xmit) {
         printf("%-22.22s ------Number  of------  Average   ----Percent  of----\n", $subject1);
         printf("%-22.22s Files %4.4s Bytes  %4.4s Xmit  Rate File %4.4s Byte %4.4s\n",
                $subject2, $dir, $dir, $dir, $dir);
         print '-' x 22,' ','-' x 10,' ','-' x 11,' ','-' x 10,' ','-' x 9,' ','-' x 9,"\n";
       }
       else {
         printf("%-32.32s ------Number  of------ -----Percent  of-----\n", $subject1);
         printf("%-32.32s Files %4.4s Bytes  %4.4s Files %4.4s Bytes %4.4s\n",
                $subject2, $dir, $dir, $dir, $dir);
         print '-' x 32,' ','-' x 10,' ','-' x 11,' ','-' x 10,' ','-' x 10,"\n";
       }
     }
   }
   if ($opt_C) {
      open(gdata, ">" . &TempFilename($filename . '.dat')) || die "$!\n";
      @labels = ();
      $lcount = 0;
   }
   $count = 0;
   $Direction = $direction;
   $Code = $code;                               # needed for ByItem sort
   $prog =<<EOF;
#   print STDERR "START main report LOOP\\n" if ($debug);
   foreach \$item ( sort ByItem keys(\%Files$VarCodes{$code}) ) {
      last if (\$opt_L && \$count >= \$opt_L);
      \@sItem = split(\$;, \$item);
      next if (\$sItem[1] ne \$direction);        # skip if wrong direction
      \$files = \$Files$VarCodes{$code}\{\$item\};
      \$bytes = \$Bytes$VarCodes{$code}\{\$item\};
      \$pctfiles = sprintf("%8.2f", 100*\$files / 
                          (\$direction eq 'o' ? $xferOFiles : $xferIFiles));
      \$pctbytes = sprintf("%8.2f", 100*\$bytes /
                          (\$direction eq 'o' ? $xferOBytes : $xferIBytes));
      \$name = \$sItem[0];
      \$name = \$DomainName{\$sItem[0]} if (\$code eq 'd' &&
                                            \$DomainName{\$sItem[0]});
      if (\$xmit) {
         \$secs  = \$Secs$VarCodes{$code}\{\$item\};
         if (\$secs < 1) { \$avgrate = 'N/A'; }
         else { \$avgrate  = sprintf("%5.1f KB/s", \$bytes/\$secs/1000); }
         if (length(\$name) > 22 ) {
	    if ( \$code eq 'z' ) { \$name = substr(\$name, 0, 21) . '+'; }
	    else { \$name = '+' . substr(\$name, -21); }
	 }
         \$HTMLline = sprintf("%-22.22s %10d %11d %10s %s  %s", \$name, \$files,
                              \$bytes, \$avgrate, \$pctfiles, \$pctbytes);
      }
      else {
         if ( length(\$name) > 32 ) { 
	    if ( \$code eq 'z' ) { \$name = substr(\$name, 0, 31) . '+'; }
	    else { \$name = '+' . substr(\$name, -31); }
	 }
         \$HTMLline = sprintf("%-32.32s %10d %11d %s %s", \$name, \$files,
                              \$bytes, \$pctfiles, \$pctbytes);
      }
      if (\$opt_C && \$lcount < $MaxData && \$lcount < 40) {
         # generate the graph data
         if (\$GraphType eq 'bar') { print gdata "\$lcount \$files \\n"; }
         else { print gdata "\$files \$lcount \\n"; }       # hbar
         \$labels[\$lcount] = \$code eq 'y' ? substr(\$name,0,6) : 
                               (\$code eq 'm' && \$lcount % 2 != 0) ? ' ' :
                               \$name ;
	 \$lcount++;
      }
      \$count++;
      if (\$opt_P) { write; }	# printed report
      elsif (\$opt_Z) {		# HTML report
         if (\$code eq 'Y') { &RefLine(\$sItem[0], \$files, \$pctfiles) }
	 else { print "\$HTMLline\\n"; }
      }
      else { print STDERR "ERROR: Unknown report format\\n"; }
   }
EOF
#   print STDERR $prog if ($debug);
   eval $prog;
   warn $@ if $@;
   print "</PRE>\n" if ($opt_Z);
#   print STDERR "After main report eval routine\n" if ($debug);
   if ($opt_C) {
      close(gdata);
      if ($code eq 'y') {
         $title2 = "$hostname - $labels[$#labels] to $labels[0]";
      }
      else { $title2 = $subtitle; }
      # generate the graph parameter (.par) file
      &PrepareGraphParameters($filename, $GraphTitle, $title2, $GraphType,
                              $MinX, $GraphType eq 'bar'  ? $lcount : $MaxX,
                              $MinY, $GraphType eq 'hbar' ? $lcount : $MaxY,
                              $XLabel, $YLabel);
      # produce graph
      &GenerateGIF($filename, $autoscale);
   }
}

#-------------------------------------------------------------------------
#	Sort subroutines
#-------------------------------------------------------------------------
sub ByDate { &CompareDate($b, $a); }

sub ByItem {
   # Item format: "Key Direction"
   # $Code must defined before sort
   ($key1) = split($;, $a);
   ($key2) = split($;, $b);
   # Sort items
   $prog =<<EOF;
   \&$SortCodes{$Code}('$key1', '$key2',
                       \$Files$VarCodes{$Code}\{\$a\},
                       \$Files$VarCodes{$Code}\{\$b\});
EOF
#   print STDERR $prog if ($debug);
   eval $prog;
}
#-------------------------------------------------------------------------
#	Comparison subroutines to support the sort routines
#-------------------------------------------------------------------------
sub CompareFileAlpha {
   # 2 level sort: by number of files, then name
   # Parameters: Name1, Name2, Files1, Files2
   if ($_[2] == $_[3]) {
      # Sort by name
      $_[0] cmp $_[1];
   }
   else {
      # Reverse Sort by number of files
      $_[3] <=> $_[2];
   }
}

sub CompareNumber {
   $_[0] <=> $_[1];
}

sub CompareWeekday {
   $WDay1 = index($Wstr, $_[0]);
   $WDay2 = index($Wstr, $_[1]);
   $WDay1 <=> $WDay2;
}

sub CompareMonth {
   $Month1 = index($Mstr, $_[0]);
   $Month2 = index($Mstr, $_[1]);
   $Month1 <=> $Month2;
}

sub CompareFileHost {
   # Sort by File # then Host name with all Numeric IP addresses last
   if ($_[2] == $_[3]) {
      # sort host name
      @Host1 = split(/\./, $_[0]);
      @Host2 = split(/\./, $_[1]);
      if (int($Host1[0]) > 0 && int($Host2[0]) > 0) {
         # Sort IP addresses
         $H1 = $Host1[0] * 2**24 + $Host1[1] * 2**16 + $Host1[2] * 256 + $Host1[3];
         $H2 = $Host2[0] * 2**24 + $Host2[1] * 2**16 + $Host2[2] * 256 + $Host2[3];
         $H1 - $H2
      }
      elsif (int($Host1[0]) > 0 || int($Host2[0]) > 0) {
         # Put Host names before IP addresses
         if (int($Host1[0]) > 0) { 1 }
         else { -1 }
      }
      else {
         # Sort host names
         $_[0] cmp $_[1]
      }
   }
   else {       # sort by number of files
      $_[3] <=> $_[2];
   }
}

sub CompareFileDomain {
   # 2 level sort: by number of files, then Domain name
   # Parameters: Domain1, Domain2, Files1, Files2
   if ($_[2] == $_[3]) {
      # Sort by Domain Name
      $DomainName{$_[0]} cmp $DomainName{$_[1]}
   }
   else {
      # Reverse Sort by number of files
      $_[3] <=> $_[2]
   }
}

sub CompareRDate { &CompareDate( $_[1], $_[0]); }

sub CompareDate {
   # Parameters: 2 date strings of format "MMM DD YYYY"
   @date1 = split(' ',$_[0]);
   @date2 = split(' ',$_[1]);
   $date1  = $date1[2] * 4800;
   $date2  = $date2[2] * 4800;
   $date1 += index($Mstr, $date1[0]) * 400/3;
   $date2 += index($Mstr, $date2[0]) * 400/3;
   $date1 += $date1[1];
   $date2 += $date2[1];
   $date1 <=> $date2;
}
#-------------------------------------------------------------------------
#	Misc. subroutines
#-------------------------------------------------------------------------
sub CreateTypeData {
   # Groupings: HTML(.html,.htm), Scripts(), Maps,
   #            Images(bmp,gif,xbm,jpg), Movies(mpg,mov,scm),
   #            Test(txt), Postscript(ps), Informational,
   #            Compressed(gz,Z,zip), Macintosh(hqx,bin,sea,sit),
   #		MS-DOS Executable(com,exe),
   #		Other Documents (doc,rtf,sam),
   #            cc:Mail Databases(#,#.*), Other(), Errors()
   # calculate the file extensions used
   foreach $file ( keys(%FilesFile) ) {
      @sfile = split($;, $file);
#      next if ($sfile[1] ne $_[0]);
      $_ = $sfile[0];
      # case insensitive matching for BBS & EMWAC
      $_ =~ tr/A-Z/a-z/ if ($opt_B || $opt_X);
      if ( /(\.(htm|html)$)|\/$/) {		# HTML file
         $FilesType{'HTML','o'}             += $FilesFile{$file};
         $BytesType{'HTML','o'}             += $BytesFile{$file};
#	 print "HTML: $_\n" if $debug;
      }
      elsif ( /\.(bmp|gif|xbm|jpg)$/) {		# Graphics - Stills
         $FilesType{'Images','o'}           += $FilesFile{$file};
         $BytesType{'Images','o'}           += $BytesFile{$file};
#	 print "IMAGE: $_\n" if $debug;
      }
      elsif ( /\.(mpg|mov|scm)$/) {		# Graphics - Movies
         $FilesType{'Movies','o'}           += $FilesFile{$file};
         $BytesType{'Movies','o'}           += $BytesFile{$file};
#	 print "MOVIE: $_\n" if $debug;
      }
      elsif ( /\?[0-9]{1,3},[0-9]{1,3}$/) {             # Image Maps
         $FilesType{'Image Maps','o'}       += $FilesFile{$file};
         $BytesType{'Image Maps','o'}       += $BytesFile{$file};
      }
      elsif ( /\/cgi-bin\// || /\?/) {                  # Scripts
         $FilesType{'Scripts','o'}          += $FilesFile{$file};
         $BytesType{'Scripts','o'}          += $BytesFile{$file};
      }
      elsif ( /\.txt$/ ) {                              # Text Files
         $FilesType{'Text Files','o'}       += $FilesFile{$file};
         $BytesType{'Text Files','o'}       += $BytesFile{$file};
#	 print "TEXT: $_\n" if $debug;
      }
      elsif ( /\.ps$/ ) {                               # Postscript Files
         $FilesType{'Postscript Files','o'} += $FilesFile{$file};
         $BytesType{'Postscript Files','o'} += $BytesFile{$file};
      }
      elsif (/\.(gz|Z|[zZ][iI][pP])$/) {	# Compressed Files
         $FilesType{'Compressed Files','o'} += $FilesFile{$file};
         $BytesType{'Compressed Files','o'} += $BytesFile{$file};
#	 print "COMPRESSED: $_\n" if $debug;
      }
      elsif (/\.(hqx|bin|sea|sit)$/) {		# Macintosh Files
         $FilesType{'Macintosh Files','o'}  += $FilesFile{$file};
         $BytesType{'Macintosh Files','o'}  += $BytesFile{$file};
      }
      elsif (/\.(com|exe)$/) {			# MS-DOS Executables
	 $FilesType{'MS-DOS Executables','o'}  += $FilesFile{$file};
         $BytesType{'MS-DOS Executables','o'}  += $BytesFile{$file};
#	 print "DOS: $_\n" if $debug;
      }
      elsif (/\.(doc|rtf|sam)$/) {	# Other Documents
         $FilesType{'Other Documents','o'}  += $FilesFile{$file};
         $BytesType{'Other Documents','o'}  += $BytesFile{$file};
      }
      elsif ( /\/\.[\w\.\-]+$/ || /\/README.*$/ ) {     # Informational files (. files)
         # Files that start with a '.' or 'README'
         $FilesType{'Informational Files','o'}  += $FilesFile{$file};
         $BytesType{'Informational Files','o'}  += $BytesFile{$file};
      }
      elsif ( /\/[0-9]{1,8}$/ || /\/[0-9]{1,8}\..{1,3}$/ ) {# cc:Mail databases ??
         $FilesType{'cc:Mail Databases ??','o'} += $FilesFile{$file};
         $BytesType{'cc:Mail Databases ??','o'} += $BytesFile{$file};
#	 print "CCMAIL: $_\n" if $debug;
      }
      else {
         $FilesType{'Unknown','o'}          += $FilesFile{$file};
         $BytesType{'Unknown','o'}          += $BytesFile{$file};
         print STDERR "Other File Type: $FilesFile{$file} of $_\n" if ($debug);
      }
   }
}

sub RefLine {
   # Parameter: Full path of file
   # Outputs a HTML line with the file name, its description
   # ( from index file ), and a link to the file
   #
   # This routine does not handle everything. Still need to improve
   #
   local( $path, $requests, $percent ) = @_;
   local( $anchor, $root, $file, $index, $desc );

   if ($opt_W) {
      $anchor = 'ftp://';
      $root = $DataRoots{'opt_W'};
   }
   elsif ($opt_G) {
      $anchor = 'gopher://';
      $root = $DataRoots{'opt_G'};
   }
   elsif ($opt_N) {
      $anchor = 'http://';
      $root = $DataRoots{'opt_N'};
   }
   $anchor .= $hostname . $path;
   $file = `basename $path`;	chop($file);
   if ( $opt_G || $opt_W ) {		# Get description - FTP & Gopher
      $index = `dirname $path`;	chop($index);
      chop($index) if ( $index =~ q#/$# );  	# substr($index, -1) eq '/');
      $index = $root . $index . "/$IndexFile";
      if ( -f $index ) {
         $desc = `grep $file $index`; chop($desc);
	 @desc = split(/\s+/,$desc);
	 $desc = join(" ",@desc[1..$#desc]);
      }
   }
   if ( $file =~ /\.html$|\.htm$/ ) {	# Get description - HTML
      $fullpath = $root . $path;
      if ( -f $fullpath ) {
         $tdesc = `grep -i "<title>" $fullpath`; chop($tdesc);
	 $start= index($tdesc,">") + 1;
	 $len  = index($tdesc,"<",$start);
	 if ( $len > $start ) {
	    $len -= $start;
	    $desc = substr($tdesc, $start, $len);
	 }
      }
   }
   if (length($desc) <= length($file)) {
      $desc = "- No Description Available";
   }
   $desc = $file . " " x (16 - length($file)) . " " . $desc;
# print ' ' x (5 - length($requests)), "$requests <A HREF=\"$anchor\">$desc</A>\n";
   print "$percent <A HREF=\"$anchor\">$desc</A>\n";
}

sub OptToDate {
   # Parameter: 1 string of format "MMDDYYYY"
   # Converts date to format "MMM DD YYYY"
   substr($Mstr, (substr($_[0],0,2)-1)*3,3) . " " . substr($_[0],2,2)
     . " " . substr($_[0],4,4)
}

sub WeekDay {
   # Parameters: Month, Day, Year, Hour, Min, Sec, GMT
   @TimeDate = @_;
   $TimeDate[0] = &MonthNum($TimeDate[0]);      # change month to number
   @TimeDate = localtime( &timelocal(@TimeDate[5,4,3,1,0,2],0,0,0));
   substr($Wstr,$TimeDate[6] * 3,3);            # return string of week day
}

sub MonthNum {
   # Parameter: 3 letter month abbreviation
   index($Mstr,$_[0]) / 3;
}

sub hostbyaddr {
   # Return the host name lookup from a numeric IP address string
   local ($dotname) = @_;
   local ($inetaddr) = pack("C4", split(/\./, $dotname, 4));
   local ($name) = gethostbyaddr($inetaddr, &AF_INET);
   if ($name) { $name; }
   else { $dotname; }
}

sub TempFilename {
  "$tmp/$userID$PID" . "_" . shift(@_);
}

sub GetConfigFilePath {
   # Get the path of where this program is located
   # and return path to configuration file at same location
   local($program) = $0;
   local(@ProgramPath) = split(m#/#, $program);
   local($dir);

   while ( -l $program ) {
      $program = readlink($program);
      if ($program =~ m#^/# ) { @ProgramPath = () }
      else { pop(@ProgramPath) }
      push(@ProgramPath, split(m#/#, $program));
      $program = join('/', @ProgramPath);
   }
   pop(@ProgramPath);
   $dir = join('/', @ProgramPath);
   if (! $dir) { $dir = '.' }
   "$dir/$ConfigFile"
}

sub GetCompany {
   # Function returns the company name
   local($host) = @_;	# Parameter is: last 2 or 3 parts of host name in reverse order
   local( $name, $aliases, $type, $len, $proto, $RemoteAddr, $count, $Company );

   $sockaddr = 'S n a4 x8';
   if (! $LocalSock ) {
      local( $hostname, $LocalAddr );
      chop($hostname = `hostname`);
      # For speed I just use the #s, instead of doing lookups
      # TCP protocol = 6, Whois port = 43
      ($name, $aliases, $proto) = getprotobyname('tcp');
#      ($name, $aliases, $port) = getservbyname('whois', 'tcp');
      ($name, $aliases, $type, $len, $LocalAddr) = gethostbyname($hostname);
      $LocalSock = pack($sockaddr, &AF_INET, 0, $LocalAddr);
   }
   socket(S, &AF_INET, &SOCK_STREAM, $proto) || die "ERROR: creating socket: $!";
   bind(S, $LocalSock) || die "ERROR: binding socket: $!";
   @host = split(/\./,$host);
   undef $found;
   foreach $DomPattern (keys (%Whois)) {
      if ($host[0] =~ /^($DomPattern)$/ ) {
	 ($WhoisServer, $Query, $OrgPattern) = split($;, $Whois{$DomPattern});
	 $found++;
	 last;
      }
   }
   return $host if ! $found;
   if ( $LastWhoisServer ne $WhoisServer ) {
      if ( $WhoisServer =~ /^[0-9\.]+$/ ) {
	 $RemoteAddr = pack("C4",split(/\./,$WhoisServer));
      }
      else {
	 ($name, $aliases, $type, $len, $RemoteAddr) = gethostbyname($WhoisServer);
      }
      $RemoteSock = pack($sockaddr, &AF_INET, 43, $RemoteAddr);
      $LastWhoisServer = $WhoisServer;
   }
   connect(S, $RemoteSock) || die "ERROR: connecting to $WhoisServer: $!";
   select(S); $| = 1; select(stdout);
   # build query
   $Query =~ s/#/join('.',reverse(@host))/e;
   # Query whois server
   print S $Query,"\r\n";
   undef $found;
   print STDERR "Pattern=$OrgPattern=\n" if $debug;
   while (<S>) {
      if ( ! $found ) {
         if ( /^No (match|entries)/ ) { $Company = $host; }
	 elsif ( /^$OrgPattern$/ ) {
	    $Company = $1;
	    $Company = join(' ',split(/\s+/,$Company));
	    $found++;
         }
#	 else { print STDERR if $debug; }
      }
   }
   close(S);
   $Company = $host if ! $Company;
#   print STDERR "Company=$Company=\n" if $debug;
   $Company;
}

#-------------------------------------------------------------------------
#	Graphing subroutines
#-------------------------------------------------------------------------

sub GenerateGIF {
  local ($name,$autoscale) = @_;
  local ($command);
  local ($par, $dat, $eps, $gif, $chip);

  # create files names
  $par  = &TempFilename($name . '.par');
  $dat  = &TempFilename($name . '.dat');
  $eps  = &TempFilename($name . '.eps');
  $gif  = &TempFilename($name . '.gif');
  $chip = &TempFilename($name . '-chip.gif');

  $command = "$graphprogram -autoscale $autoscale -device 2 -eps " .
             "-printfile $eps -noask -p $par $dat";

  system("$command");
  system("$graphconvertor $eps $gif");
  system("$graphconvertor -geometry $csize! -colors $ccolor $eps $chip");
  system("mv $gif $destination/$name" . '.gif');
  system("mv $chip $destination/$name" . '-chip.gif');
  system("rm $eps $dat $par") unless ($debug);
}

sub PrepareGraphParameters {
  local($name, $title, $subtitle, $type,
        $xmin, $xmax, $ymin, $ymax,
        $x_label, $y_label) = @_;

  local($xTickMajor,$xTickMinor,
        $xTickLabelSize,$xTickLabelType,$xTickSpec,
        $yTickMajor,$yTickMinor,
        $yTickLabelSize,$yTickLabelType,$yTickSpec,$yTickMajorEnable,
        $xvmin, $xvmax);

  # variables used to produce the parameter file:
  #     $type           graph type (xy, bar, hbar, stackedbar ,stackedhbar,
  #                                 logx, logy, logxy)
  #     $xmin           scaling limits of the graph
  #     $xmax               "
  #     $ymin               "
  #     $ymax               "
  #     $title          graph title
  #     $subtitle       graph subtitle
  #     $outlinec       color of lines
  #     $fillc          color of fill
  #     $x_label        label of X axis
  #     $y_label        label of Y axis
  #     $bgc            background color
  #     @labels         tick labels
  #     $date           date & time of graph creation
  #     $Version        iisstat version string

  if ($type eq 'bar') {
    ($xTickMajor, $xTickMinor, $xTickLabelSize, $xTickLabelType, $xTickSpec,
     $yTickMajor, $yTickMinor,
     $yTickLabelSize, $yTickLabelType, $yTickSpec, $yTickMajorEnable,
     $xvmin, $xvmax) =
        (1, 0.5, 0.77, 'spec', $#labels + 1, 1000, 500, 1.0, 'auto', 0, 'on', 0.15, 0.9);
  }
  elsif ($type eq 'hbar') {
    ($xTickMajor, $xTickMinor, $xTickLabelSize, $xTickLabelType, $xTickSpec,
     $yTickMajor, $yTickMinor,
     $yTickLabelSize, $yTickLabelType, $yTickSpec, $yTickMajorEnable,
     $xvmin, $xvmax) =
        (200, 100, 1.0, 'auto', 0, 1, 1, 0.79, 'spec', $#labels + 1, 'off', 0.35, 0.9);
  }
  else { die "ERROR: unsupported graph type: $type"; }

  $textColor = 1;
  $font = 4;
  $date = `date`; chop($date);

  open(parfile, ">" . &TempFilename($name . '.par'));
  print parfile <<end_of_part1;
# ACE/gr parameter file
#
#
default linestyle 1
default linewidth 1
default color 1
default char size 1.0
default font $font
default font source 0
default symbol size 1.0
with string 
    string on
    string loctype view
    string g0
    string 0.950, 0.0840
    string font 2
    string char size 0.60
    string linewidth 1
    string color 6
    string rot 0
    string just 1
string def "Created on $date by $Version"
with g0
g0 on
g0 label off
g0 hidden false
g0 type $type
g0 autoscale type AUTO
g0 fixedpoint off
g0 fixedpoint type 0
g0 fixedpoint xy 0.000000, 0.000000
g0 fixedpoint format general general
g0 fixedpoint prec 6, 6
    world xmin $xmin
    world xmax $xmax
    world ymin $ymin
    world ymax $ymax
    view xmin $xvmin
    view xmax $xvmax
    view ymin 0.15
    view ymax 0.8
    title "$title"
    title font $font
    title size 1.500000
    title color $textColor
    title linewidth 1
    subtitle "$subtitle"
    subtitle font $font
    subtitle size 1.000000
    subtitle color $textColor
    subtitle linewidth 1
    s0 type xy
    s0 symbol 0
    s0 symbol size 1.000000
    s0 symbol fill 0
    s0 symbol color 1
    s0 symbol linewidth 1
    s0 symbol linestyle 1
    s0 symbol center false
    s0 symbol char 0
    s0 skip 0
    s0 linestyle 1
    s0 linewidth 1
    s0 color $outlinec
    s0 fill 1
    s0 fill with color
    s0 fill color $fillc
    s0 fill pattern 0
    s0 errorbar type BOTH
    s0 errorbar length 1.000000
    s0 errorbar linewidth 1
    s0 errorbar linestyle 1
    s0 errorbar riser on
    s0 errorbar riser linewidth 1
    s0 errorbar riser linestyle 1
    s0 xyz 0.000000, 0.000000
    s0 comment "comment"
    xaxis  label "$x_label"
    xaxis  label layout para
    xaxis  label place auto
    xaxis  label char size 1.000000
    xaxis  label font $font
    xaxis  label color $textColor
    xaxis  label linewidth 1
    xaxis  bar off
    xaxis  bar color 1
    xaxis  bar linestyle 1
    xaxis  bar linewidth 1
    xaxis  ticklabel on
    xaxis  ticklabel type $xTickLabelType
    xaxis  ticklabel prec 1
    xaxis  ticklabel format decimal
    xaxis  ticklabel append ""
    xaxis  ticklabel prepend ""
    xaxis  ticklabel layout horizontal
    xaxis  ticklabel skip 0
    xaxis  ticklabel stagger 0
    xaxis  ticklabel op bottom
    xaxis  ticklabel sign normal
    xaxis  ticklabel start type $xTickLabelType
    xaxis  ticklabel start 0.000000
    xaxis  ticklabel stop type auto
    xaxis  ticklabel stop 0.000000
    xaxis  ticklabel char size $xTickLabelSize
    xaxis  ticklabel font $font
    xaxis  ticklabel color $textColor
    xaxis  ticklabel linewidth 1
    xaxis  tick on
    xaxis  tick major $xTickMajor
    xaxis  tick major on
    xaxis  tick major color 1
    xaxis  tick major linewidth 1
    xaxis  tick major linestyle 1
    xaxis  tick major grid off
    xaxis  tick minor $xTickMinor
    xaxis  tick minor on
    xaxis  tick minor color 1
    xaxis  tick minor linewidth 1
    xaxis  tick minor linestyle 1
    xaxis  tick minor size 0.500000
    xaxis  tick minor grid off
    xaxis  tick offsetx 0.000000
    xaxis  tick offsety 0.000000
    xaxis  tick default 6
    xaxis  tick in
    xaxis  tick log off
    xaxis  tick size 1.000000
    xaxis  tick op bottom
    xaxis  tick type auto
    xaxis  tick spec $xTickSpec
end_of_part1
  if ($type eq 'bar') {
    for ($i = 0; $i <= $#labels; $i++) {
	print parfile "    xaxis  tick $i, $i\n";
	print parfile "    xaxis  ticklabel $i, \"" .  $labels[$i] . "\"\n";}
  }
  print parfile <<"end_of_part2";
    yaxis  label "$y_label"
    yaxis  label layout para
    yaxis  label place auto
    yaxis  label char size 1.000000
    yaxis  label font $font
    yaxis  label color $textColor
    yaxis  label linewidth 1
    yaxis  bar off
    yaxis  bar color 1
    yaxis  bar linestyle 1
    yaxis  bar linewidth 1
    yaxis  ticklabel on
    yaxis  ticklabel type $yTickLabelType
    yaxis  ticklabel prec 1
    yaxis  ticklabel format decimal
    yaxis  ticklabel append ""
    yaxis  ticklabel prepend ""
    yaxis  ticklabel layout horizontal
    yaxis  ticklabel skip 0
    yaxis  ticklabel stagger 0
    yaxis  ticklabel op left
    yaxis  ticklabel sign normal
    yaxis  ticklabel start type $yTickLabelType
    yaxis  ticklabel start 0.000000
    yaxis  ticklabel stop type auto
    yaxis  ticklabel stop 0.000000
    yaxis  ticklabel char size $yTickLabelSize
    yaxis  ticklabel font $font
    yaxis  ticklabel color $textColor
    yaxis  ticklabel linewidth 1
    yaxis  tick on
    yaxis  tick major $yTickMajor
    yaxis  tick major $yTickMajorEnable
    yaxis  tick major color 1
    yaxis  tick major linewidth 1
    yaxis  tick major linestyle 1
    yaxis  tick major grid off
    yaxis  tick minor $yTickMinor
    yaxis  tick minor on
    yaxis  tick minor color 1
    yaxis  tick minor linewidth 1
    yaxis  tick minor linestyle 1
    yaxis  tick minor size 0.500000
    yaxis  tick minor grid off
    yaxis  tick offsetx 0.000000
    yaxis  tick offsety 0.000000
    yaxis  tick default 6
    yaxis  tick in
    yaxis  tick log off
    yaxis  tick size 1.000000
    yaxis  tick op both
    yaxis  tick type auto
    yaxis  tick spec $yTickSpec
end_of_part2
  if ($type eq 'hbar') {
    for ($i = 0; $i <= $#labels; $i++) {
	print parfile "    yaxis  tick $i, $i\n";
	print parfile "    yaxis  ticklabel $i, \"" .  $labels[$i] . "\"\n";}
  }
  print parfile <<"end_of_part3";
    legend off
    legend loctype view
    legend layout 0
    legend vgap 2
    legend hgap 1
    legend length 4
    legend box off
    legend box fill off
    legend box fill with color
    legend box fill color 0
    legend box fill pattern 1
    legend box color 1
    legend box linewidth 1
    legend box linestyle 1
    legend x1 0.8
    legend y1 0.8
    legend font $font
    legend char size 1.000000
    legend linestyle 1
    legend linewidth 1
    legend color $textColor
    frame on
    frame type 0
    frame linestyle 1
    frame linewidth 1
    frame color 1
    frame fill on
    frame background color $bgc
end_of_part3
  close parfile;
}
